from collections import namedtuple
from datetime import datetime, timedelta
from functools import update_wrapper
from typing import Optional, Generator, Any, Callable, Tuple, List

import click

from wanikani_api.client import Client as WaniKaniClient
from wanikani_notifier.notifiers import notifier
from wanikani_notifier.notifiers.notifier import Notifier
from wanikani_notifier.notifiers.console import ConsoleNotifier
from wanikani_notifier.notifiers.pushover import PushoverNotifier
from wanikani_notifier.notifiers.pushsafer import PushSaferNotifier
from wanikani_notifier.wanikani import get_notification_message, get_available_assignments


def processor(f: Callable):
    def new_func(*args, **kwargs):
        def processor(context: Context, message_stream: Generator[str, Any, None]):
            return f(context, message_stream, *args, **kwargs)

        return processor

    return update_wrapper(new_func, f)


def generator(f: Callable):
    @processor
    def new_func(context: Context, message_stream: str, *args, **kwargs) -> Generator[str, Any, None]:
        yield from message_stream
        yield from f(context, *args, **kwargs)

    return update_wrapper(new_func, f)


@click.group(chain=True)
@click.option("--wanikani", required=True, help="WaniKani API token")
@click.option("--stop-if-empty/--continue-even-empty",
              default=True,
              help="Determines whether chaining occurs when one command does not trigger a new message to notify"
              )
def cli(wanikani: str, stop_if_empty: bool):
    pass  # pragma: nocover


Context = namedtuple("Context", ("wanikani_client", "stop_if_empty"))


@cli.resultcallback()
def process_all(processors, wanikani: str, stop_if_empty: bool):
    context = Context(wanikani_client=WaniKaniClient(wanikani), stop_if_empty=stop_if_empty)

    message_stream = ()
    for processor in processors:
        message_stream = processor(context, message_stream)

    for _ in message_stream:
        pass


@cli.command("available_assignments_now")
@click.option(
    "--since",
    required=False,
    default=-1,
    help="How many hours since assignments are accounted for (-1 meaning forever)",
    show_default=True
)
@click.option(
    "--min",
    "min_assignments",
    required=False,
    type=click.IntRange(min=1),
    default=1,
    help="Minimum number of assignments to generate a message",
    show_default=True
)
@generator
def cli_available_assignments_now(context: Context, since: int, min_assignments: int):
    yield available_assignments_now(context.wanikani_client, since, min_assignments)


def available_assignments_now(wanikani_client: WaniKaniClient, since: int, min_assignments: int):
    current_time_rounded = datetime.utcnow()
    start_time = (current_time_rounded - (timedelta(hours=since) - timedelta(seconds=1)) if since >= 0 else None)
    assignments_available_now = get_available_assignments(wanikani_client,
                                                          start=start_time,
                                                          end=current_time_rounded
                                                          )
    if sum(assignments_available_now) >= min_assignments:
        return get_notification_message(assignments_available_now, message_template="{} are now available!")
    else:
        return


@cli.command("all_available_assignments")
@generator
def cli_all_available_assignments(context: Context):
    yield all_available_assignments(context.wanikani_client)


def all_available_assignments(wanikani_client: WaniKaniClient):
    current_time_rounded = datetime.utcnow()
    all_assignments_available = get_available_assignments(wanikani_client, end=current_time_rounded)
    return get_notification_message(all_assignments_available, message_template="In total, there are {} to do.")


@cli.command("notify")
@click.option("--console/--no-console", required=False, help="Activates notifications though the console")
@click.option("--pushsafer",
              required=False,
              help="Activates notifications though PushSafer by providing the private key"
              )
@click.option("--pushover",
              nargs=2,
              type=str,
              required=False,
              help="Activates notifications though Pushover by providing the app key and the user key"
              )
@processor
def cli_notify(context: Context,
               message_stream: Generator[str, Any, None],
               pushsafer: Optional[str],
               pushover: Optional[Tuple[str, str]],
               console: Optional[bool]
               ) -> None:
    notifiers = []
    if pushsafer:
        notifiers.append(notifier.factory.create(PushSaferNotifier.key(), private_key=pushsafer))
    if pushover:
        notifiers.append(notifier.factory.create(PushoverNotifier.key(), app_token=pushover[0], user_token=pushover[1]))
    if console:
        notifiers.append(notifier.factory.create(ConsoleNotifier.key()))

    messages = list(message_stream)
    if context.stop_if_empty and not all(messages):
        yield
        return

    notify("\n".join(m for m in messages if m), notifiers)
    yield


def notify(message: str, notifiers: List[Notifier]) -> None:
    if message:
        for n in notifiers:
            n.notify(title="WaniKani", message=message, url="https://www.wanikani.com/dashboard")



import datetime
from unittest.mock import MagicMock

import pytest
from wanikani_api.client import Client as WaniKaniClient
from pytest_mock import MockerFixture

from wanikani_notifier.wanikani import get_available_assignments, get_notification_message, AvailableAssignments


class MockedLesson:
    def __init__(self):
        self.unlocked_at = True
        self.started_at = False


class MockedReview:
    def __init__(self):
        self.unlocked_at = True
        self.started_at = True


@pytest.fixture
def mocked_wk_client(mocker: MockerFixture) -> MagicMock:
    return mocker.Mock(spec=WaniKaniClient)


def test_available_assignments(mocked_wk_client):
    mocked_wk_client.assignments.return_value = []

    get_available_assignments(mocked_wk_client, datetime.datetime.utcnow())

    mocked_wk_client.assignments.assert_called()


def test_get_message_no_new_reviews_no_new_lessons():
    message = get_notification_message(AvailableAssignments(reviews=0, lessons=0))

    assert message is None


def test_get_message_new_reviews_no_new_lessons():
    message = get_notification_message(AvailableAssignments(reviews=2, lessons=0))

    assert "lesson" not in message
    assert "review" in message


def test_get_message_no_new_reviews_new_lessons():
    message = get_notification_message(AvailableAssignments(reviews=0, lessons=3))

    assert "lesson" in message
    assert "review" not in message


def test_get_message_new_reviews_new_lessons():
    message = get_notification_message(AvailableAssignments(reviews=2, lessons=3))

    assert "lesson" in message
    assert "review" in message



from abc import ABC, abstractmethod
from typing import Any, Dict

import pytest

from wanikani_notifier.notifiers.notifier import NoMessageProvided, Notifier, factory


class NotifierTester(ABC):
    @abstractmethod
    @pytest.fixture
    def imp(self) -> Notifier:
        pass

    @classmethod
    @abstractmethod
    def key(cls) -> str:
        pass

    @abstractmethod
    @pytest.fixture
    def build_parameters(self) -> Dict[str, Any]:
        pass

    def test_builder(self, build_parameters):
        assert factory.create(self.key, **build_parameters) is not None

    def test_notify_none_message(self, imp):
        with pytest.raises(NoMessageProvided):
            imp.notify("title", None)

    def test_notify_empty_message(self, imp):
        with pytest.raises(NoMessageProvided):
            imp.notify("title", "")

    def test_notify_simple_message(self, imp):
        imp.notify("title", "message")

    def test_notify_message_all_options(self, imp):
        imp.notify("title", "message", "url", "icon")



from typing import Optional

import chump

from wanikani_notifier.notifiers.notifier import Notifier, NoMessageProvided, factory


class PushoverNotifier(Notifier):
    def __init__(self, app_token: str, user_token: str):
        self._app = chump.Application(app_token)
        self._user = self._app.get_user(user_token)

    @classmethod
    def key(cls) -> str:
        return "pushover"

    @classmethod
    def build(cls, app_token: str, user_token: str, **_ignored) -> "PushoverNotifier":
        return PushoverNotifier(app_token, user_token)

    def notify(self, title: str, message: str, url: Optional[str] = None, icon: Optional[str] = None):
        if not message:
            raise NoMessageProvided

        self._user.send_message(title=title, message=message, url=url)


factory.register_builder(PushoverNotifier.key(), PushoverNotifier.build)



from datetime import datetime
from collections import namedtuple
from typing import Optional

import pytz
from wanikani_api.client import Client as WaniKaniClient

AvailableAssignments = namedtuple("AvailableAssignments", ("reviews", "lessons"))


def get_available_assignments(wanikani_client: WaniKaniClient,
                              end: datetime,
                              start: datetime = None
                              ) -> AvailableAssignments:
    """
    Gets the number of reviews and lessons that are available in the provided time period.

    :rtype: object
    :param wanikani_client: WaniKani client to use for fetching assignments
    :param start: Start of the time period when assignments are considered (inclusive).
    :param end: End of the time period when assignments are considered (inclusive).
    :return: the available assignments.
    """
    user_level = wanikani_client.user_information().level
    review_count = sum(1
                       for a in wanikani_client.assignments(fetch_all=True, unlocked=True, started=True,
                                                            available_after=start, available_before=end)
                       if a.subject.level <= user_level
                       )
    lesson_count = sum(1
                       for a in wanikani_client.assignments(fetch_all=True, unlocked=True, started=False)
                       if a.created_at.replace(tzinfo=pytz.utc) <= end.replace(tzinfo=pytz.utc)
                       and (start.replace(tzinfo=pytz.utc) <= a.created_at.replace(tzinfo=pytz.utc) if start else True)
                       )

    return AvailableAssignments(reviews=review_count, lessons=lesson_count)


def get_notification_message(available_assignments: AvailableAssignments,
                             message_template: Optional[str] = None) -> Optional[str]:
    """
    Creates the notification message corresponding to the available assignments.

    :param available_assignments: Number of reviews and lessons that are available.
    :param message_template: Message template where the generated number of reviews and lessons
                                replaces the {} format specifier
    :return: a string containing the notification message to send if at least one assignment is available,
                None otherwise
    """
    if not sum(available_assignments):
        return

    available_lessons_message = ""
    available_reviews_message = ""
    if available_assignments.lessons:
        available_lessons_message = f"{available_assignments.lessons} lessons"
    if available_assignments.reviews:
        available_reviews_message = f"{available_assignments.reviews} reviews"

    full_message = " and ".join(m for m in [available_lessons_message, available_reviews_message] if m)
    return message_template.format(full_message) if message_template else full_message



from typing import Tuple
from unittest.mock import MagicMock

import pytest
from click.testing import CliRunner
from pytest_mock import MockerFixture

from wanikani_notifier.cli import cli
from wanikani_notifier.cli import notify, available_assignments_now, all_available_assignments
from wanikani_notifier.wanikani import AvailableAssignments


###################
#     Fixtures    #
###################


@pytest.fixture
def mocked_wk_client(mocker: MockerFixture) -> MagicMock:
    return mocker.patch("wanikani_notifier.cli.WaniKaniClient")


@pytest.fixture
def mocked_notifier_creator(mocker: MockerFixture):
    return mocker.patch("wanikani_notifier.notifiers.notifier.factory.create")


@pytest.fixture
def mocked_get_available_assignments(mocker: MockerFixture) -> MagicMock:
    return mocker.patch("wanikani_notifier.cli.get_available_assignments")


@pytest.fixture
def mocked_available_assignments_now(mocker: MockerFixture) -> MagicMock:
    return mocker.patch("wanikani_notifier.cli.available_assignments_now")


@pytest.fixture
def mocked_all_available_assignments(mocker: MockerFixture) -> MagicMock:
    return mocker.patch("wanikani_notifier.cli.all_available_assignments")


###################
#    CLI tests    #
###################

class TestCli:
    def test_cli_no_commands_no_options(self):
        runner = CliRunner()
        result = runner.invoke(cli, "")

        assert "Usage" in result.stdout

    def test_cli_no_commands_with_wanikani_token(self, mocked_wk_client):
        runner = CliRunner()
        result = runner.invoke(cli, "--wanikani __TOKEN__")

        assert result.exit_code == 2

    def test_cli_available_assignments_now_without_options(self, mocked_wk_client):
        runner = CliRunner()
        result = runner.invoke(cli, "--wanikani __TOKEN__ available_assignments_now")

        assert result.exit_code == 0

    def test_cli_available_assignments_now_with_options(self, mocked_wk_client):
        runner = CliRunner()
        result = runner.invoke(cli, "--wanikani __TOKEN__ available_assignments_now --since 1")

        assert result.exit_code == 0

    def test_cli_available_assigments(self, mocked_wk_client):
        runner = CliRunner()
        result = runner.invoke(cli, "--wanikani __TOKEN__ all_available_assignments")

        assert result.exit_code == 0

    def test_cli_notify_no_notifiers(self, mocked_notifier_creator):
        runner = CliRunner()
        result = runner.invoke(cli, "--wanikani __TOKEN__ notify")

        assert result.exit_code == 0
        mocked_notifier_creator.assert_not_called()

    def test_cli_notify_all_notifiers(self, mocked_notifier_creator):
        runner = CliRunner()
        result = runner.invoke(cli,
                               """
                               --wanikani __TOKEN__
                               notify --pushsafer __TOKEN__ --pushover __APP_TOKEN__ __USER_TOKEN__ --console
                               """
                               )

        assert result.exit_code == 0
        assert mocked_notifier_creator.call_count == 3


###################
#  CLI use cases  #
###################
class TestUseCases:
    @staticmethod
    @pytest.fixture(params=[True, False], ids=["stop", "continue"])
    def notify_new_and_all_use_case(request) -> Tuple[str, bool]:
        return (
            """
            --wanikani __TOKEN__ {}
            available_assignments_now
            all_available_assignments
            notify --console
            """.format(("--continue-even-empty" if not request.param else "")),
            request.param
        )

    @pytest.mark.parametrize("now_available,all_available", [
        pytest.param(None, None, id="none"),
        pytest.param("__MESSAGE__", None, id="some_new_none_all"),
        pytest.param(None, "__MESSAGE__", id="none_new_some_all"),
        pytest.param("__MESSAGE__", "__MESSAGE__", id="some_new_some_all"),
    ])
    def test_cli_notify_new_and_all_use_case(self,
                                             mocked_notifier_creator,
                                             mocked_available_assignments_now,
                                             mocked_all_available_assignments,
                                             notify_new_and_all_use_case,
                                             now_available,
                                             all_available
                                             ):
        mocked_available_assignments_now.return_value = now_available
        mocked_all_available_assignments.return_value = all_available
        if not notify_new_and_all_use_case[1]:
            expect_notify = any((now_available, all_available))
        else:
            expect_notify = all((now_available, all_available))

        runner = CliRunner()
        result = runner.invoke(cli, notify_new_and_all_use_case[0])

        assert result.exit_code == 0
        assert mocked_notifier_creator.call_count == 1
        assert mocked_available_assignments_now.call_count == 1
        assert mocked_all_available_assignments.call_count == 1
        assert mocked_notifier_creator.return_value.notify.call_count == (1 if expect_notify else 0)


###################
#  Command tests  #
###################
class TestCommands:
    @pytest.mark.parametrize("assignments,min_assignments,expected_message",
                             [
                                 pytest.param(AvailableAssignments(0, 2),
                                              1,
                                              "2 lessons are now available!",
                                              id="lessons_only"),
                                 pytest.param(AvailableAssignments(3, 0),
                                              1,
                                              "3 reviews are now available!",
                                              id="reviews_only"),
                                 pytest.param(AvailableAssignments(4, 5),
                                              1,
                                              "5 lessons and 4 reviews are now available!",
                                              id="lessons_and_reviews"),
                                 pytest.param(AvailableAssignments(4, 5),
                                              10,
                                              None,
                                              id="lessons_and_reviews_below_min"),
                                 pytest.param(AvailableAssignments(4, 5),
                                              9,
                                              "5 lessons and 4 reviews are now available!",
                                              id="lessons_and_reviews_equal_min"),
                                 pytest.param(AvailableAssignments(0, 0),
                                              1,
                                              None,
                                              id="none")
                             ]
                             )
    def test_available_assignments_now(self, mocked_get_available_assignments,
                                       assignments, min_assignments, expected_message):
        mocked_get_available_assignments.return_value = assignments
        message = available_assignments_now(None, 0, min_assignments)

        assert message == expected_message

    @pytest.mark.parametrize("assignments,expected_message",
                             [
                                 pytest.param(AvailableAssignments(0, 2),
                                              "In total, there are 2 lessons to do.",
                                              id="lessons_only"),
                                 pytest.param(AvailableAssignments(3, 0),
                                              "In total, there are 3 reviews to do.",
                                              id="reviews_only"),
                                 pytest.param(AvailableAssignments(4, 5),
                                              "In total, there are 5 lessons and 4 reviews to do.",
                                              id="lessons_and_reviews"),
                                 pytest.param(AvailableAssignments(0, 0),
                                              None,
                                              id="none")
                             ]
                             )
    def test_all_available_assignments(self, mocked_get_available_assignments, assignments, expected_message):
        mocked_get_available_assignments.return_value = assignments
        message = all_available_assignments(None)

        assert message == expected_message

    def test_notify_no_notifiers_no_messages(self, mocked_notifier_creator):
        notify("", [])

        assert mocked_notifier_creator.call_count == 0
        assert mocked_notifier_creator.return_value.notify.call_count == 0

    def test_notify_no_notifiers_some_messages(self, mocked_notifier_creator):
        notify("some message", [])

        assert mocked_notifier_creator.call_count == 0
        assert mocked_notifier_creator.return_value.notify.call_count == 0

    def test_notify_all_notifiers_no_messages(self, mocked_notifier_creator):
        notify("", [mocked_notifier_creator(), mocked_notifier_creator(), mocked_notifier_creator()])

        assert mocked_notifier_creator.call_count == 3
        assert mocked_notifier_creator.return_value.notify.call_count == 0

    def test_notify_all_notifiers_some_messages(self, mocked_notifier_creator):
        notifiers = [mocked_notifier_creator(), mocked_notifier_creator(), mocked_notifier_creator()]

        notify("some message", notifiers)

        assert mocked_notifier_creator.call_count == 3
        assert mocked_notifier_creator.return_value.notify.call_count == 3



#!/usr/bin/env python

"""The setup script."""

import pip
from setuptools import setup, find_packages

try: # for pip >= 10
    from pip._internal.req import parse_requirements
except ImportError: # for pip <= 9.0.3
    from pip.req import parse_requirements


with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()


## workaround derived from: https://github.com/pypa/pip/issues/7645#issuecomment-578210649
parsed_requirements = parse_requirements(
    'requirements/prod.txt',
    session='workaround'
)

parsed_test_requirements = parse_requirements(
    'requirements/test.txt',
    session='workaround'
)

try:
    requirements = [str(ir.req) for ir in parsed_requirements]
    test_requirements = [str(tr.req) for tr in parsed_test_requirements]
except AttributeError:
    requirements = [str(ir.requirement) for ir in parsed_requirements]
    test_requirements = [str(tr.requirement) for tr in parsed_test_requirements]

setup(
    author="Laurent Radoux",
    author_email='radoux.laurent@gmail.com',
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    description="Notifies a WaniKani user for new reviews to do.",
    entry_points={
        'console_scripts': [
            'wanikani_notifier=wanikani_notifier.cli:cli',
        ],
    },
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='wanikani_notifier',
    name='wanikani_notifier',
    packages=find_packages(include=['wanikani_notifier', 'wanikani_notifier.*']),
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/laurent-radoux/wanikani_notifier',
    version='0.5.2',
    zip_safe=False,
)



from typing import Dict, Any
from unittest.mock import MagicMock

import pytest
from pytest_mock import MockerFixture

from tests.notifiers.test_notifier import NotifierTester
from wanikani_notifier.notifiers.notifier import Notifier
from wanikani_notifier.notifiers.pushsafer import PushSaferNotifier


@pytest.fixture
def mocked_pushsafer_init(mocker: MockerFixture) -> MagicMock:
    mocked = mocker.patch("pushsafer.Client")
    yield mocked
    mocked.assert_called_once()


@pytest.fixture
def mocked_pushsafer_send_message(mocked_pushsafer_init) -> MagicMock:
    return mocked_pushsafer_init.return_value.send_message


class TestPushSaferNotifier(NotifierTester):
    @property
    def key(self) -> str:
        return "pushsafer"

    @pytest.fixture
    def build_parameters(self) -> Dict[str, Any]:
        return {"private_key": "__TOKEN__"}

    @pytest.fixture
    def imp(self, mocked_pushsafer_init) -> Notifier:
        yield PushSaferNotifier("__TOKEN__")

    def test_notify_simple_message(self, imp, mocked_pushsafer_send_message):
        super().test_notify_simple_message(imp)
        mocked_pushsafer_send_message.assert_called()

    def test_notify_message_all_options(self, imp, mocked_pushsafer_send_message):
        super().test_notify_message_all_options(imp)
        mocked_pushsafer_send_message.assert_called()
